package net.knarcraft.bookswithoutborders.command;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import org.bukkit.Material;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;

import static net.knarcraft.bookswithoutborders.BooksWithoutBorders.bookPriceQuantity;
import static net.knarcraft.bookswithoutborders.BooksWithoutBorders.bookPriceType;
import static net.knarcraft.bookswithoutborders.BooksWithoutBorders.sendErrorMessage;
import static net.knarcraft.bookswithoutborders.BooksWithoutBorders.sendSuccessMessage;
import static net.knarcraft.bookswithoutborders.BooksWithoutBordersSettings.getCommandColor;
import static net.knarcraft.bookswithoutborders.BooksWithoutBordersSettings.getSuccessColor;

/**
 * Command executor for the books without borders (bwb) command
 */
public class CommandBooksWithoutBorders implements CommandExecutor {

    private final BooksWithoutBorders booksWithoutBorders;

    public CommandBooksWithoutBorders(BooksWithoutBorders booksWithoutBorders) {
        this.booksWithoutBorders = booksWithoutBorders;
    }

    @Override
    public boolean onCommand(CommandSender sender, Command command, String label, String[] args) {
        if (sender instanceof Player) {
            showPlayerCommands(sender);
        } else {
            showConsoleCommands(sender);
        }
        return true;
    }

    /**
     * Shows all commands available to the console
     *
     * @param sender <p>The console which sent the command</p>
     */
    private void showConsoleCommands(CommandSender sender) {
        sender.sendMessage(getCommandColor() + "Use: /bwb [Command]");
        sender.sendMessage(getCommandColor() + "[] denote parameters");
        sender.sendMessage(getCommandColor() + "Commands:");
        sender.sendMessage(getCommandColor() + "\nReload:" + getSuccessColor() + " Reloads BwB's config file");
        sender.sendMessage(getCommandColor() + "givePublic [file name or number] [player name] [true/false]: " + getSuccessColor());
        sendSuccessMessage(sender, "Gives the selected player a book from the public directory");
        sendSuccessMessage(sender, "If no file is specified, a list of available files is returned");
        sender.sendMessage(getCommandColor() + "deletePublic [file name or number]: " + getSuccessColor() + "Deletes the specified");
        sendSuccessMessage(sender, "file in the public directory");
        sendSuccessMessage(sender, "If no file is specified, a list of available files is returned");
    }

    /**
     * Shows all commands available to the sending player
     *
     * @param sender <p>The player which sent the command</p>
     */
    private void showPlayerCommands(CommandSender sender) {
        //Lists all commands
        sender.sendMessage(getCommandColor() + "Use: /bwb [Command]");
        sender.sendMessage(getCommandColor() + "[] denote parameters");

        if (booksWithoutBorders.booksHavePrice()) {
            if (bookPriceType != Material.AIR) {
                sendErrorMessage(sender, "[" + (int) bookPriceQuantity + " " + bookPriceType.toString() + "(s) are required to create a book]");
            } else {
                sendErrorMessage(sender, "[" + BooksWithoutBorders.eco.format(bookPriceQuantity) + " is required to create a book]");
            }
        }
        sender.sendMessage(getCommandColor() + "Commands:");

        if (sender.hasPermission("bookswithoutborders.load")) {
            sender.sendMessage(getCommandColor() + "\nLoad [file name or number] [# of copies] [true/false]:");
            sendSuccessMessage(sender, "Creates a book from the specified file and gives it to the player");
            sendSuccessMessage(sender, "If no file is specified, a list of available files is returned");
            sendSuccessMessage(sender, "If true is specified the book will be signed, if false it will be");
            sendSuccessMessage(sender, "unsigned");
        }
        if (sender.hasPermission("bookswithoutborders.loadPublic")) {
            sender.sendMessage(getCommandColor() + "loadPublic [file name or number] [# of copies] [true/false]:");
            sendSuccessMessage(sender, "Same as Load, but views files in the public directory");
        }
        if (sender.hasPermission("bookswithoutborders.save")) {
            sender.sendMessage("\n" + getCommandColor() + "Save [true/false]: " + getSuccessColor() + "Saves the book the player is");
            sendSuccessMessage(sender, "holding to a text file in a private directory");
            sendSuccessMessage(sender, "If true is specified, a book of the same name by the same");
            sendSuccessMessage(sender, "author will be overwritten by the new book");
        }
        if (sender.hasPermission("bookswithoutborders.savePublic")) {
            sender.sendMessage(getCommandColor() + "savePublic [true/false]: " + getSuccessColor() + "Same as Save,");
            sendSuccessMessage(sender, "but saves files in the public directory");
        }
        if (sender.hasPermission("bookswithoutborders.give")) {
            sender.sendMessage("\n" + getCommandColor() + "Give [file name or number] [player name] [# of copies] [true/false]:");
            sendSuccessMessage(sender, "Gives the selected player a book from your personal directory");
        }
        if (sender.hasPermission("bookswithoutborders.givePublic")) {
            sender.sendMessage(getCommandColor() + "givePublic [file name or number] [player name] [# of copies] [true/false]:");
            sendSuccessMessage(sender, "Same as give, but uses books from the public directory");
        }
        if (sender.hasPermission("bookswithoutborders.delete")) {
            sender.sendMessage(getCommandColor() + "\nDelete [file name or number]: " + getSuccessColor() + "Deletes the specified");
            sendSuccessMessage(sender, "file in the player's directory");
            sendSuccessMessage(sender, "If no file is specified, a list of available files is returned");
        }
        if (sender.hasPermission("bookswithoutborders.admin")) {
            sender.sendMessage(getCommandColor() + "deletePublic [file name or number]: " + getSuccessColor() + "Same as Delete,");
            sendSuccessMessage(sender, "but deletes files in the public directory");
            sender.sendMessage(getCommandColor() + "\nReload:" + getSuccessColor() + " Reloads BwB's configuration file");
        }

        if (sender.hasPermission("bookswithoutborders.unsign")) {
            sender.sendMessage("\n" + getCommandColor() + "Unsign: " + getSuccessColor() + "Un-signs the book the player is holding");
        }
        if (sender.hasPermission("bookswithoutborders.copy")) {
            sender.sendMessage("\n" + getCommandColor() + "Copy [number of copies]: " + getSuccessColor() + "Copies the book the player is holding");
        }

        if (sender.hasPermission("bookswithoutborders.encrypt")) {
            sender.sendMessage("\n" + getCommandColor() + "Encrypt [key] [style]: " + getSuccessColor() + "Encrypts the book the player is holding");
            sender.sendMessage(getCommandColor() + "[key]" + getSuccessColor() + " is required and can be any phrase or number excluding spaces");
            sender.sendMessage(getCommandColor() + "[style]" + getSuccessColor() + " is not required. Possible values are \"DNA\" or \"Magic\"");
        }
        if (sender.hasPermission("bookswithoutborders.groupEncrypt")) {
            sender.sendMessage("\n" + getCommandColor() + "groupEncrypt [group name] [key] [style]: " + getSuccessColor() + "Encrypts book so that only players with the" +
                    "\n bookswithoutborders.decrypt." + getCommandColor() + "[group name]" + getSuccessColor() + " permission may decrypt the book by holding and left clicking the book");
        }
        if (sender.hasPermission("bookswithoutborders.decrypt")) {
            sender.sendMessage("\n" + getCommandColor() + "Decrypt [key]: " + getSuccessColor() + "Decrypts the book the player is holding");
            sender.sendMessage(getCommandColor() + "[key]" + getSuccessColor() + " is required and MUST be IDENTICAL to the key used to encrypt held book");
        }
        if (sender.hasPermission("bookswithoutborders.setTitle")) {
            sender.sendMessage("\n" + getCommandColor() + "setTitle [title]: " + getSuccessColor() + "Sets the title of the book/item the player is holding");
        }
        if (sender.hasPermission("bookswithoutborders.setAuthor")) {
            sender.sendMessage("\n" + getCommandColor() + "setAuthor [author]: " + getSuccessColor() + "Sets the author of the book the player is holding");
        }
        if (sender.hasPermission("bookswithoutborders.setLore")) {
            sender.sendMessage("\n" + getCommandColor() + "setLore [lore]: " + getSuccessColor() + "Sets the lore of the item the player is holding");
            sendSuccessMessage(sender, "Insert the lore_line_separator character to force a new line\n[\"~\" by default]");
        }
        if (sender.hasPermission("bookswithoutborders.setBookPrice")) {
            sender.sendMessage("\n" + getCommandColor() + "setBookPrice [Item/Eco] [quantity]: " + getSuccessColor() + "Sets the per-book-price to create a book via commands." +
                    "\nIf [Item], the item in the player's hand in the amount of [quantity] will be the price.\nIf [Eco], a Vault based economy will be used for price." +
                    "\nIf neither [Item/Eco] or [quantity] are specified the current price to create books will be removed.");
        }
    }

}
