package net.knarcraft.bookswithoutborders.command;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import net.knarcraft.bookswithoutborders.state.ItemSlot;
import net.knarcraft.bookswithoutborders.utility.InventoryHelper;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;

/**
 * Command executor for the save command
 */
public class CommandSave implements CommandExecutor {

    private final BooksWithoutBorders booksWithoutBorders;

    public CommandSave(BooksWithoutBorders booksWithoutBorders) {
        this.booksWithoutBorders = booksWithoutBorders;
    }

    @Override
    public boolean onCommand(CommandSender sender, Command command, String label, String[] args) {
        return saveHeldBook(sender, args, false);
    }

    /**
     * Saves the player's held book if it exists
     *
     * @param sender     <p>The sender of the command</p>
     * @param args       <p>The arguments given</p>
     * @param savePublic <p>Whether to save the book in the public directory or the player directory</p>
     * @return <p>True if a book was saved successfully</p>
     */
    boolean saveHeldBook(CommandSender sender, String[] args, boolean savePublic) {
        if (!(sender instanceof Player player)) {
            BooksWithoutBorders.sendErrorMessage(sender, "This command can only be used by a player!");
            return false;
        }

        ItemSlot holdingSlot = InventoryHelper.getHeldSlotBook(player, false, false, false, false);
        if (holdingSlot != ItemSlot.NONE) {
            ItemStack holdingItem = InventoryHelper.getHeldItem(player, holdingSlot == ItemSlot.MAIN_HAND);
            boolean duplicate = args.length == 1 && Boolean.parseBoolean(args[0]);
            booksWithoutBorders.saveBook(player, holdingItem, duplicate, savePublic);
            return true;
        } else {
            BooksWithoutBorders.sendErrorMessage(sender, "You must be holding a written book or book and quill to save it!");
            return false;
        }
    }
}
